import os
from functools import partial
from glob import glob
from pathlib import Path

from qtpy.QtCore import QPoint, Qt, Signal
from qtpy.QtGui import QDragEnterEvent, QDropEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.custom_load_dialog import SelectDirectoryDialog
from PartSeg.common_gui.main_window import OPEN_DIRECTORY
from PartSegCore.analysis.calculation_plan import MaskMapper

IO_BATCH_DIRECTORY = "io.batch_directory"


class AcceptFiles(QDialog):
    def __init__(self, files):
        super().__init__()
        self.ok = QPushButton("Add", self)
        self.ok.clicked.connect(self.accept)
        discard = QPushButton("Discard", self)
        discard.clicked.connect(self.close)
        self.files = QListWidget(self)
        self.files.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        for file_name in files:
            self.files.addItem(file_name)
        for i in range(self.files.count()):
            self.files.item(i).setSelected(True)
        self.ok.setDefault(True)
        self.ok.setAutoDefault(True)

        layout = QVBoxLayout()
        layout.addWidget(QLabel(f"Found {len(files)} files"))
        layout.addWidget(self.files)
        butt_layout = QHBoxLayout()
        butt_layout.addWidget(discard)
        butt_layout.addStretch()
        butt_layout.addWidget(self.ok)
        layout.addLayout(butt_layout)
        self.setLayout(layout)

    def selection_changed(self):
        if self.files.selectedItems().count() == 0:
            self.ok.setDisabled(True)
        else:
            self.ok.setEnabled(True)

    def get_files(self):
        return [str(item.text()) for item in self.files.selectedItems()]


class FileListItem(QListWidgetItem):
    def __init__(self, file_path):
        size = os.stat(file_path).st_size
        size = float(size) / (1024**2)
        super().__init__(f"{file_path:s} ({size:.2f} MB)")
        self.setTextAlignment(Qt.AlignmentFlag.AlignRight)
        self.file_path = file_path


class FileListWidget(QListWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)


class AddFiles(QWidget):
    """Docstring for AddFiles."""

    file_list_changed = Signal(set)

    def __init__(self, settings: BaseSettings, parent=None, btn_layout=QHBoxLayout):
        """TODO: to be defined1."""
        super().__init__(parent)
        self.mask_list: list[MaskMapper] = []
        self.settings = settings
        self.files_to_proceed = set()
        self.paths_input = QLineEdit(self)
        self.add_suffix_btn = QPushButton("*[0-9].tif", self)
        self.add_suffix_btn.clicked.connect(self.add_suffix_action)
        self.selected_files = FileListWidget(self)
        self.selected_files.itemSelectionChanged.connect(self.file_chosen)
        self.found_button = QPushButton("Find all", self)
        self.found_button.clicked.connect(self.find_all)
        self.select_files_button = QPushButton("Select files")
        self.select_dir_button = QPushButton("Select directory")
        self.select_files_button.clicked.connect(self.select_files)
        self.select_dir_button.clicked.connect(self.select_directory)
        self.delete_button = QPushButton("Remove file", self)
        self.delete_button.setDisabled(True)
        self.delete_button.clicked.connect(self.delete_element)
        self.clean_button = QPushButton("Remove all", self)
        self.clean_button.clicked.connect(self.clean)
        layout = QVBoxLayout()
        input_layout = QHBoxLayout()
        input_layout.addWidget(self.paths_input)
        input_layout.addWidget(self.add_suffix_btn)
        layout.addLayout(input_layout)
        select_layout = btn_layout()
        select_layout.addWidget(self.select_files_button)
        select_layout.addWidget(self.select_dir_button)
        select_layout.addWidget(self.found_button)
        select_layout.addStretch()
        select_layout.addWidget(self.clean_button)
        select_layout.addWidget(self.delete_button)
        layout.addLayout(select_layout)
        layout.addWidget(self.selected_files)
        self.setLayout(layout)
        self.setAcceptDrops(True)
        self.selected_files.customContextMenuRequested.connect(self.files_context_menu)

    def add_suffix_action(self):
        if self.paths_input.text() and not self.paths_input.text().endswith(self.add_suffix_btn.text()):
            self.paths_input.setText(os.path.join(self.paths_input.text(), self.add_suffix_btn.text()))

    def files_context_menu(self, point: QPoint):
        element = self.selected_files.itemAt(point)
        if element is None:
            return
        menu = QMenu()
        menu.addAction("Load image").triggered.connect(self._load_file)
        for mask_def in self.mask_list:
            menu.addAction(f"Load with mask '{mask_def.name}'").triggered.connect(
                partial(self._load_file_with_mask, mask_def)
            )
        menu.addAction("Delete").triggered.connect(self.delete_element)
        menu.exec_(self.selected_files.mapToGlobal(point))

    def _load_file(self):
        file_path = self.selected_files.item(self.selected_files.currentRow()).file_path
        self.settings._load_files_call([file_path])  # pylint: disable=protected-access

    def _load_file_with_mask(self, mask_mapper: MaskMapper):
        file_path = self.selected_files.item(self.selected_files.currentRow()).file_path
        mask_path = mask_mapper.get_mask_path(file_path)
        self.settings._load_files_call([file_path, mask_path])  # pylint: disable=protected-access

    def dragEnterEvent(self, event: QDragEnterEvent):  # pylint: disable=no-self-use
        if event.mimeData().hasFormat("text/plain"):
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        files_list = event.mimeData().text().split("\n")
        self.parse_drop_file_list(files_list)

    def parse_drop_file_list(self, files_list):
        res_list = []
        base_path = self.paths_input.text()
        is_broken_path = "*" in base_path or not os.path.isdir(base_path)

        for file_path in files_list:
            if os.path.isabs(file_path):
                res_list.append(file_path)
            elif is_broken_path:
                QMessageBox.warning(self, "Not proper path", f"Current path {base_path} is not a proper directory.")
                return
            else:
                res_list.append(os.path.join(base_path, file_path.strip()))
        if missed_files := [x for x in res_list if not os.path.exists(x)]:
            if len(missed_files) > 6:
                missed_files = missed_files[:6] + ["..."]
            missed_files_str = "<br>".join(missed_files)
            QMessageBox().warning(
                self,
                "Missed Files",
                f"Cannot find files:<br>{missed_files_str}<br>Set proper base directory using <i>Select directory</i>",
            )
        else:
            self.update_files_list(res_list)

    def find_all(self):
        # When it fails again then there is a need to provide test
        # with input that causes crash
        paths = glob(str(self.paths_input.text()))

        paths = sorted(x for x in (set(paths) - self.files_to_proceed) if not os.path.isdir(x))
        if len(paths) > 0:
            self.update_files_list(paths)

        else:
            QMessageBox.warning(self, "No new files", "No new files found", QMessageBox.StandardButton.Ok)

    def update_files_list(self, paths):
        dialog = AcceptFiles(paths)
        if dialog.exec_():
            new_paths = dialog.get_files()
            for path in new_paths:
                self.selected_files.addItem(FileListItem(path))
            self.files_to_proceed.update(new_paths)
            self.file_list_changed.emit(self.files_to_proceed)

    def select_files(self):
        dial = QFileDialog(self, "Select files")
        dial.setDirectory(self.settings.get(IO_BATCH_DIRECTORY, self.settings.get(OPEN_DIRECTORY, str(Path.home()))))
        dial.setFileMode(QFileDialog.FileMode.ExistingFiles)
        if dial.exec_():
            self.settings.set(IO_BATCH_DIRECTORY, os.path.dirname(str(dial.selectedFiles()[0])))
            new_paths = sorted(set(map(str, dial.selectedFiles())) - self.files_to_proceed)
            for path in new_paths:
                self.selected_files.addItem(FileListItem(path))
            self.files_to_proceed.update(new_paths)
            self.file_list_changed.emit(self.files_to_proceed)

    def select_directory(self):
        dial = SelectDirectoryDialog(
            settings=self.settings,
            settings_path=[IO_BATCH_DIRECTORY, OPEN_DIRECTORY],
            default_directory=None,
            parent=self,
        )
        if dial.exec_():
            self.paths_input.setText(dial.selectedFiles()[0])

    def file_chosen(self):
        self.delete_button.setEnabled(True)

    def delete_element(self):
        item = self.selected_files.takeItem(self.selected_files.currentRow())
        self.files_to_proceed.remove(item.file_path)
        self.file_list_changed.emit(self.files_to_proceed)
        if self.selected_files.count() == 0:
            self.delete_button.setDisabled(True)

    def clean(self):
        self.selected_files.clear()
        self.files_to_proceed.clear()
        self.file_list_changed.emit(self.files_to_proceed)

    def get_paths(self):
        return sorted(self.files_to_proceed)



import contextlib
import gc
import json
from importlib.metadata import version
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest
from local_migrator import object_hook
from napari.layers import Image as NapariImage
from napari.layers import Labels
from napari.utils import Colormap
from packaging.version import parse as parse_version
from qtpy.QtCore import QObject, QTimer, Signal

from PartSeg._roi_analysis.profile_export import ExportDialog, ImportDialog
from PartSeg.common_gui.custom_load_dialog import CustomLoadDialog
from PartSeg.common_gui.custom_save_dialog import CustomSaveDialog
from PartSeg.common_gui.napari_image_view import SearchType
from PartSeg.plugins.napari_widgets import (
    CopyLabelsWidget,
    ImageColormap,
    LayerMetadata,
    MaskCreate,
    ROIAnalysisExtraction,
    ROIMaskExtraction,
    SearchLabel,
    _settings,
)
from PartSeg.plugins.napari_widgets.algorithm_widgets import (
    BorderSmoothingModel,
    CompareType,
    ConnectedComponentsModel,
    DoubleThresholdModel,
    FlowType,
    NoiseFilterModel,
    SplitCoreObjectsModel,
    Threshold,
    ThresholdModel,
    WatershedModel,
)
from PartSeg.plugins.napari_widgets.colormap_control import NapariColormapControl
from PartSeg.plugins.napari_widgets.lables_control import LabelSelector, NapariLabelShow
from PartSeg.plugins.napari_widgets.measurement_widget import update_properties
from PartSeg.plugins.napari_widgets.roi_extraction_algorithms import ProfilePreviewDialog, QInputDialog
from PartSeg.plugins.napari_widgets.search_label_widget import HIGHLIGHT_LABEL_NAME
from PartSegCore import Units
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.load_functions import LoadProfileFromJSON
from PartSegCore.analysis.measurement_calculation import Volume, Voxels
from PartSegCore.analysis.save_functions import SaveProfilesToJSON
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection
from PartSegCore.segmentation import ROIExtractionResult
from PartSegCore.segmentation.border_smoothing import SmoothAlgorithmSelection
from PartSegCore.segmentation.noise_filtering import NoiseFilterSelection
from PartSegCore.segmentation.threshold import DoubleThresholdSelection, ThresholdSelection
from PartSegCore.segmentation.watershed import WatershedSelection
from PartSegCore.universal_const import LayerNamingFormat

NAPARI_GE_5_0 = parse_version(version("napari")) >= parse_version("0.5.0a1")
NAPARI_GE_4_19 = parse_version(version("napari")) >= parse_version("0.4.19a1")

if NAPARI_GE_4_19:

    def check_auto_mode(layer):
        from napari.utils.colormaps import CyclicLabelColormap

        assert isinstance(layer.colormap, CyclicLabelColormap)

    def check_direct_mode(layer):
        from napari.utils.colormaps import DirectLabelColormap

        assert isinstance(layer.colormap, DirectLabelColormap)

else:

    def check_auto_mode(layer):
        assert layer.color_mode == "auto"

    def check_direct_mode(layer):
        assert layer.color_mode == "direct"


@pytest.fixture(autouse=True)
def _clean_settings(tmp_path):
    old_settings = _settings._SETTINGS
    _settings._SETTINGS = _settings.PartSegNapariSettings(tmp_path)
    yield
    _settings._SETTINGS = old_settings


def no_action(*_):  # skipcq: PTC-W0049
    pass


def test_dummy_for_gc():
    gc.collect()


@pytest.mark.parametrize("widget_class", [ROIAnalysisExtraction, ROIMaskExtraction])
def test_extraction_widget(make_napari_viewer, widget_class, monkeypatch, qtbot):
    viewer = make_napari_viewer()
    viewer.add_image(np.ones((10, 10)))
    widget = widget_class(napari_viewer=viewer)
    viewer.window.add_dock_widget(widget)
    widget.reset_choices()
    widget.update_image()
    viewer.add_image(np.ones((10, 10)))
    monkeypatch.setattr(QInputDialog, "getText", get_text_mock("prof3"))
    monkeypatch.setattr(ProfilePreviewDialog, "exec_", no_action)
    assert widget.profile_combo_box.count() == 1
    widget.save_action()
    with qtbot.waitSignal(widget.algorithm_chose.algorithm_choose.currentIndexChanged):
        widget.algorithm_chose.algorithm_choose.setCurrentIndex(1)
    with qtbot.waitSignal(widget.profile_combo_box.currentIndexChanged):
        widget.profile_combo_box.setCurrentIndex(1)
    assert widget.profile_combo_box.count() == 2
    widget.refresh_profiles()
    assert widget.profile_combo_box.count() == 2
    widget.manage_action()
    widget.select_profile("prof3")
    widget.mask_name = "1"
    widget.update_image()
    assert not widget.mask_name
    # check if mask is not cleaned if channels does not change
    widget.mask_name = "1"
    widget.update_image()
    assert widget.mask_name == "1"

    res = ROIExtractionResult(
        np.ones((10, 10), dtype=np.uint8), widget.profile_dict["prof3"], roi_annotation={1: {"foo": 1, "baz": 2}}
    )
    res.roi[0, 0] = 2
    res.roi[1, 1] = 0

    widget.set_result(res)
    widget.set_result(res)

    curr_widget = widget.algorithm_chose.current_widget()
    curr_widget.form_widget.widgets_dict["mask"]._widget.value = viewer.layers["ROI"]
    widget.update_mask()


def get_text_mock(text):
    def func(*_args, **_kwargs):
        return text, True

    return func


@pytest.mark.parametrize("register", [AnalysisAlgorithmSelection, MaskAlgorithmSelection])
def test_profile_preview_dialog(part_settings, register, qtbot, monkeypatch, tmp_path):
    alg = register.get_default()
    profiles = {
        "prof1": ROIExtractionProfile(name="prof1", algorithm=alg.name, values=alg.values),
        "prof2": ROIExtractionProfile(name="prof2", algorithm=alg.name, values=alg.values),
    }
    dialog = ProfilePreviewDialog(profiles, register, part_settings)
    qtbot.add_widget(dialog)
    assert dialog.profile_view.toPlainText() == ""
    assert dialog.profile_list.count() == 2
    with qtbot.waitSignal(dialog.profile_list.currentTextChanged):
        dialog.profile_list.setCurrentRow(0)
    assert dialog.profile_list.currentItem().text() == "prof1"
    assert dialog.profile_view.toPlainText() != ""
    monkeypatch.setattr(QInputDialog, "getText", get_text_mock("prof3"))
    monkeypatch.setattr(CustomSaveDialog, "exec_", lambda x: True)
    monkeypatch.setattr(
        CustomSaveDialog,
        "get_result",
        lambda x: (tmp_path / "profile.json", None, SaveProfilesToJSON, SaveProfilesToJSON.get_default_values()),
    )
    monkeypatch.setattr(ExportDialog, "exec_", lambda x: True)
    monkeypatch.setattr(ExportDialog, "get_export_list", lambda x: ["prof1"])
    dialog.export_action()
    dialog.rename_action()
    assert "prof3" in profiles
    assert len(profiles) == 2
    with qtbot.waitSignal(dialog.profile_list.currentTextChanged):
        dialog.profile_list.setCurrentRow(0)
    assert dialog.profile_list.currentItem().text() == "prof2"

    dialog.delete_action()
    assert len(profiles) == 1
    with qtbot.waitSignal(dialog.profile_list.currentTextChanged):
        dialog.profile_list.setCurrentRow(0)

    monkeypatch.setattr(ImportDialog, "exec_", lambda x: True)
    monkeypatch.setattr(ImportDialog, "get_import_list", lambda x: [("prof1", "prof1")])
    monkeypatch.setattr(CustomLoadDialog, "exec_", lambda x: True)
    monkeypatch.setattr(
        CustomLoadDialog, "get_result", lambda x: ([tmp_path / "profile.json"], None, LoadProfileFromJSON)
    )

    dialog.import_action()
    assert len(profiles) == 2
    assert dialog.profile_list.count() == 2


def test_simple_measurement_create(make_napari_viewer, qtbot):
    from PartSeg.plugins.napari_widgets.simple_measurement_widget import SimpleMeasurement

    data = np.zeros((10, 10), dtype=np.uint8)

    viewer = make_napari_viewer()
    viewer.add_labels(data, name="label")
    measurement = SimpleMeasurement(viewer)
    viewer.window.add_dock_widget(measurement)
    measurement.reset_choices()
    measurement.labels_choice.reset_choices()
    measurement.labels_choice.value = viewer.layers["label"]
    for el in measurement.measurement_layout:
        if el.text in [Volume.get_name(), Voxels.get_name()]:
            el.value = True

    measurement._calculate()
    for _ in range(10):
        qtbot.wait(200)
        if measurement.calculate_btn.enabled:
            break

    assert measurement.calculate_btn.enabled


@pytest.mark.enablethread
@pytest.mark.enabledialog
@pytest.mark.usefixtures("qtbot")
def test_measurement_create(make_napari_viewer, bundle_test_dir, monkeypatch):
    from PartSeg.plugins.napari_widgets.measurement_widget import Measurement

    monkeypatch.setattr(
        "PartSeg.plugins.napari_widgets.measurement_widget.show_info",
        Mock(side_effect=RuntimeError("should not be called")),
    )

    data = np.zeros((10, 10), dtype=np.uint8)
    data[2:5, 2:-2] = 1
    data[5:-2, 2:-2] = 2

    viewer = make_napari_viewer()
    viewer.add_labels(data, name="label")
    viewer.add_image(data, name="image")
    viewer.add_image(data, name="image2")
    measurement = Measurement(viewer)
    viewer.window.add_dock_widget(measurement)
    measurement.reset_choices()
    measurement_data = measurement.settings.load_metadata(str(bundle_test_dir / "napari_measurements_profile.json"))
    measurement.settings.measurement_profiles["test"] = measurement_data["test"]
    assert measurement.measurement_widget.measurement_type.count() == 2
    measurement.measurement_widget.measurement_type.setCurrentIndex(1)
    assert measurement.measurement_widget.measurement_type.currentText() == "test"
    assert measurement.measurement_widget.recalculate_button.isEnabled()
    assert measurement.measurement_widget.check_if_measurement_can_be_calculated("test") == "test"
    assert measurement.measurement_widget.info_field.rowCount() == 0
    assert measurement.measurement_widget.info_field.columnCount() == 3
    measurement.measurement_widget.append_measurement_result()
    assert measurement.measurement_widget.info_field.rowCount() == 8
    assert measurement.measurement_widget.info_field.columnCount() == 3


def test_update_properties():
    data = np.zeros((10, 10), dtype=np.uint8)
    data[2:5, 2:-2] = 1
    data[5:-2, 2:-2] = 2
    labels = Labels(data)
    df = pd.DataFrame([[0, 0], [1, 1]], columns=["x", "y"])
    df2 = pd.DataFrame([[0, 0], [1, 1]], columns=["x", "z"])
    update_properties(df, labels, True)
    assert len(labels.properties) == 2
    assert "x" in labels.properties
    assert "y" in labels.properties
    assert np.all(labels.properties["x"] == np.array([0, 1]))
    update_properties(df2, labels, False)
    assert len(labels.properties) == 3
    assert "z" in labels.properties
    update_properties(df2, labels, True)
    assert len(labels.properties) == 2
    assert "y" not in labels.properties


def test_mask_create(make_napari_viewer, qtbot):
    data = np.zeros((10, 10), dtype=np.uint8)

    viewer = make_napari_viewer()
    viewer.add_labels(data, name="label")
    mask_create = MaskCreate(viewer)
    viewer.window.add_dock_widget(mask_create)
    mask_create.reset_choices()
    mask_create.create_mask()
    assert mask_create.mask_widget.get_dilate_radius() == 0

    assert "Mask" in viewer.layers


@pytest.fixture
def _shutdown_timers(monkeypatch):
    register = []
    old_start = QTimer.start

    def mock_start(self, interval=None):
        register.append(self)
        return old_start(self) if interval is None else old_start(self, interval)

    monkeypatch.setattr(QTimer, "start", mock_start)
    yield

    for timer in register:
        with contextlib.suppress(RuntimeError):
            timer.stop()

    for timer in register:
        with contextlib.suppress(RuntimeError):
            assert not timer.isActive()


@pytest.mark.enablethread
@pytest.mark.usefixtures("_shutdown_timers")
def test_search_labels(make_napari_viewer, qtbot):
    viewer = make_napari_viewer()
    data = np.zeros((10, 10), dtype=np.uint8)
    data[2:5, 2:-2] = 1
    data[5:-2, 2:-2] = 2
    viewer.add_labels(data, name="label")
    search = SearchLabel(napari_viewer=viewer)
    viewer.window.add_dock_widget(search)

    search.search_type.value = SearchType.Highlight
    search.component_selector.value = 1
    assert HIGHLIGHT_LABEL_NAME in viewer.layers
    qtbot.wait(500)
    assert HIGHLIGHT_LABEL_NAME in viewer.layers
    search._stop()
    assert HIGHLIGHT_LABEL_NAME not in viewer.layers
    search.search_type.value = SearchType.Zoom_in
    search.search_type.value = SearchType.Highlight
    assert HIGHLIGHT_LABEL_NAME in viewer.layers
    search.search_type.value = SearchType.Zoom_in
    assert HIGHLIGHT_LABEL_NAME not in viewer.layers
    search.search_type.value = SearchType.Highlight
    search.component_selector.value = 2


@pytest.fixture
def viewer_with_data(make_napari_viewer):
    viewer = make_napari_viewer()
    data = np.zeros((10, 10), dtype=np.uint8)
    data[2:5, 2:-2] = 1
    data[5:-2, 2:-2] = 2
    viewer.add_image(data, name="image")
    viewer.add_labels(data, name="label")
    viewer.layers.selection.clear()
    return viewer


def test_label_control(viewer_with_data, qtbot):
    widget = LabelSelector(viewer_with_data)
    qtbot.addWidget(widget)
    widget.label_view.refresh()

    assert widget.label_view.layout().count() == 2  # labels and stretch

    viewer_with_data.layers.selection.add(viewer_with_data.layers["label"])
    assert widget.label_view.layout().itemAt(0).widget().apply_label_btn.isEnabled()
    viewer_with_data.layers.selection.clear()
    viewer_with_data.layers.selection.add(viewer_with_data.layers["image"])
    assert not widget.label_view.layout().itemAt(0).widget().apply_label_btn.isEnabled()


def test_image_colormap(viewer_with_data, qtbot):
    widget = ImageColormap(viewer_with_data)
    qtbot.addWidget(widget)
    widget.colormap_list.refresh()

    viewer_with_data.layers.selection.add(viewer_with_data.layers["label"])
    assert not widget.colormap_list.grid_layout.itemAt(0).widget().apply_colormap_btn.isEnabled()
    viewer_with_data.layers.selection.clear()
    viewer_with_data.layers.selection.add(viewer_with_data.layers["image"])
    assert widget.colormap_list.grid_layout.itemAt(0).widget().apply_colormap_btn.isEnabled()


def test_napari_label_show(viewer_with_data, qtbot):
    widget = NapariLabelShow(viewer_with_data, "label_name", [[128, 234, 123], [240, 0, 0]], False)
    qtbot.addWidget(widget)
    viewer_with_data.layers.selection.add(viewer_with_data.layers["image"])
    assert not widget.apply_label_btn.isEnabled()
    viewer_with_data.layers.selection.add(viewer_with_data.layers["label"])
    assert not widget.apply_label_btn.isEnabled()
    viewer_with_data.layers.selection.remove(viewer_with_data.layers["image"])
    assert widget.apply_label_btn.isEnabled()
    check_auto_mode(viewer_with_data.layers["label"])
    with qtbot.waitSignal(widget.apply_label_btn.clicked):
        widget.apply_label_btn.click()
    check_direct_mode(viewer_with_data.layers["label"])


def test_napari_colormap_control(viewer_with_data, qtbot):
    widget = NapariColormapControl(viewer_with_data, Colormap([[0, 0, 0, 0], [1, 0, 0, 1]]), False, "image_name")
    qtbot.addWidget(widget)
    viewer_with_data.layers.selection.add(viewer_with_data.layers["label"])
    assert not widget.apply_colormap_btn.isEnabled()
    viewer_with_data.layers.selection.add(viewer_with_data.layers["image"])
    assert not widget.apply_colormap_btn.isEnabled()
    viewer_with_data.layers.selection.remove(viewer_with_data.layers["label"])
    assert widget.apply_colormap_btn.isEnabled()
    assert viewer_with_data.layers["image"].colormap.name == "gray"
    widget.apply_colormap_btn.click()
    assert viewer_with_data.layers["image"].colormap.name == "custom"


@pytest.fixture(params=range(2, 5))
def ndim_param(request):
    return request.param


@pytest.fixture
def napari_image(ndim_param):
    shape = (1,) * max(ndim_param - 3, 0) + (10,) * min(ndim_param, 3)
    slice_ = (slice(None),) * max(ndim_param - 3, 0) + (slice(2, 8),) * min(ndim_param, 3)
    data = np.zeros(shape, dtype=np.uint16)
    data[slice_] = 10000
    data[(0,) * (ndim_param - 2) + (slice(4, 6), slice(4, 6))] = 20000
    return NapariImage(data)


@pytest.fixture
def napari_labels(ndim_param):
    shape = (1,) * max(ndim_param - 3, 0) + (10,) * min(ndim_param, 3)
    slice_ = (slice(None),) * max(ndim_param - 3, 0) + (slice(2, 8),) * min(ndim_param, 3)
    data = np.zeros(shape, dtype=np.uint8)
    data[slice_] = 1
    data[..., -3, -3] = 0
    data[..., 5, 5] = 2
    return Labels(data)


@pytest.mark.parametrize("algorithm", ThresholdSelection.__register__.values())
def test_threshold_model(algorithm, napari_image):
    if algorithm.get_name() == "Kittler Illingworth":
        pytest.skip("Kittler Illingworth is not working properly")
    alg_params = ThresholdSelection(
        name=algorithm.get_name(),
        values=algorithm.__argument_class__(),
    )
    model = ThresholdModel(threshold=alg_params, data=napari_image)
    assert model.run_calculation()["layer_type"] == "labels"


@pytest.mark.parametrize("algorithm", DoubleThresholdSelection.__register__.values())
def test_double_threshold_model(algorithm, napari_image):
    alg_params = DoubleThresholdSelection(
        name=algorithm.get_name(),
        values=algorithm.__argument_class__(),
    )
    model = DoubleThresholdModel(threshold=alg_params, data=napari_image)
    assert model.run_calculation()["layer_type"] == "labels"


@pytest.mark.parametrize("algorithm", NoiseFilterSelection.__register__.values())
def test_noise_filter_model(algorithm, napari_image):
    alg_params = NoiseFilterSelection(
        name=algorithm.get_name(),
        values=algorithm.__argument_class__(),
    )
    model = NoiseFilterModel(noise_filtering=alg_params, data=napari_image)
    assert model.run_calculation()["layer_type"] == "image"


@pytest.mark.parametrize("algorithm", SmoothAlgorithmSelection.__register__.values())
def test_border_smoothing_model(algorithm, napari_labels):
    alg_params = SmoothAlgorithmSelection(
        name=algorithm.get_name(),
        values=algorithm.__argument_class__(),
    )
    model = BorderSmoothingModel(border_smoothing=alg_params, data=napari_labels)
    assert model.run_calculation()["layer_type"] == "labels"


@pytest.mark.parametrize("algorithm", WatershedSelection.__register__.values())
@pytest.mark.parametrize("operator", FlowType.__members__.values())
def test_watershed_model(algorithm, napari_image, napari_labels, operator):
    alg_params = WatershedSelection(
        name=algorithm.get_name(),
        values=algorithm.__argument_class__(),
    )
    model = WatershedModel(
        watershed=alg_params,
        data=napari_image,
        flow_area=napari_labels,
        core_objects=napari_labels,
        operator=operator,
    )
    assert model.run_calculation()["layer_type"] == "labels"


@patch("PartSeg.plugins.napari_widgets.algorithm_widgets.show_info")
def test_threshold_widget(show_patch, make_napari_viewer, qtbot, napari_image):
    viewer = make_napari_viewer()
    viewer.add_layer(napari_image)
    widget = Threshold(viewer)
    viewer.window.add_dock_widget(widget, area="right")
    widget.reset_choices()
    assert len(viewer.layers) == 1
    widget.run_operation()
    assert len(viewer.layers) == 2
    assert "Threshold labels" in viewer.layers
    widget.run_operation()
    assert len(viewer.layers) == 2
    del viewer.layers[napari_image.name]
    assert len(viewer.layers) == 1
    widget.run_operation()
    assert len(viewer.layers) == 1
    assert show_patch.called
    new_image = NapariImage(
        np.reshape(napari_image.data, (1, *napari_image.data.shape)), scale=(1, *napari_image.scale)
    )
    viewer.add_layer(new_image)
    widget.run_operation()
    assert len(viewer.layers) == 3


def test_part_seg_napari_encoder(napari_image):
    thr = ThresholdModel(data=napari_image)
    res_str = json.dumps(dict(thr), cls=_settings.PartSegNapariEncoder)
    res = json.loads(res_str, object_hook=object_hook)
    assert res["data"] == napari_image.name


def test_connected_components_model(napari_labels):
    model = ConnectedComponentsModel(data=napari_labels)
    assert model.run_calculation()["layer_type"] == "labels"


def test_split_core_objects_model(napari_labels):
    model = SplitCoreObjectsModel(data=napari_labels)
    assert model.run_calculation()["layer_type"] == "labels"


@pytest.fixture
def napari_labels2():
    data = np.zeros((10, 10, 10), dtype=np.uint8)
    data[5, 2:-2, 2:5] = 1
    data[5, 2:-2, 5:-2] = 2
    return Labels(data)


class MockDebouncer(QObject):
    triggered = Signal()

    def setTimeout(self, timeout):
        """to satisfy QSignalDebouncer interface"""

    def throttle(self, *args):
        self.triggered.emit()


@pytest.fixture
def copy_labels(make_napari_viewer, qtbot, napari_labels2, monkeypatch):
    monkeypatch.setattr("PartSeg.plugins.napari_widgets.copy_labels.QSignalDebouncer", MockDebouncer)
    viewer = make_napari_viewer()
    viewer.add_labels(napari_labels2.data)
    widget = CopyLabelsWidget(viewer)
    qtbot.addWidget(widget)
    viewer.add_layer(napari_labels2)
    return widget


class TestCopyLabels:
    def test_refresh_checkbox(self, copy_labels, napari_labels2):
        assert copy_labels.checkbox_layout.count() == 2
        copy_labels.checkbox_layout.itemAt(0).widget().setChecked(True)
        napari_labels2.selected_label = 3
        napari_labels2.data[4, 1, 1] = 3
        napari_labels2.events.set_data(value=napari_labels2.data)
        assert copy_labels.checkbox_layout.count() == 3

    def test_copy_action(self, copy_labels, napari_labels2):
        copy_labels.viewer.dims.current_step = (5, 0, 0)
        copy_labels.checkbox_layout.itemAt(0).widget().setChecked(True)
        copy_labels.lower.setValue(3)
        copy_labels.upper.setValue(7)
        copy_labels.copy_action()
        assert napari_labels2.data[4, 2, 2] == 1
        assert napari_labels2.data[4, 2, 6] == 0
        copy_labels.checkbox_layout.itemAt(0).widget().setChecked(False)
        copy_labels.lower.setValue(2)
        napari_labels2.selected_label = 2
        copy_labels.copy_action()
        assert napari_labels2.data[4, 2, 6] == 2
        assert napari_labels2.data[2, 2, 6] == 2
        assert napari_labels2.data[2, 2, 2] == 0
        copy_labels.viewer.layers.selection.clear()
        copy_labels.copy_action()

    def test_check_uncheck(self, copy_labels, qtbot):
        assert copy_labels.checkbox_layout.count() == 2
        assert not copy_labels.checkbox_layout.itemAt(0).widget().isChecked()
        copy_labels._check_all()
        assert copy_labels.checkbox_layout.itemAt(0).widget().isChecked()
        copy_labels._uncheck_all()
        assert not copy_labels.checkbox_layout.itemAt(0).widget().isChecked()


class TestLayerMetadata:
    def test_init(self, make_napari_viewer, qtbot):
        viewer = make_napari_viewer()
        widget = LayerMetadata(viewer)
        qtbot.addWidget(widget)
        assert widget._dict_viewer._data == {}

    def test_init_with_layer(self, make_napari_viewer, qtbot):
        viewer = make_napari_viewer()
        viewer.add_image(
            np.ones((10, 10)),
            contrast_limits=[0, 1],
            metadata={"foo": "bar"},
        )
        widget = LayerMetadata(viewer)
        viewer.window.add_dock_widget(widget)
        widget.reset_choices()
        assert widget.layer_selector.value is not None
        assert widget._dict_viewer._data == {"foo": "bar"}

    def test_add_layer_post_init(self, make_napari_viewer, qtbot):
        viewer = make_napari_viewer()
        widget = LayerMetadata(viewer)
        viewer.window.add_dock_widget(widget)
        assert widget._dict_viewer._data == {}
        viewer.add_image(
            np.ones((10, 10)),
            contrast_limits=[0, 1],
            metadata={"foo": "bar"},
        )
        widget.reset_choices()
        assert widget.layer_selector.value is not None
        assert widget._dict_viewer._data == {"foo": "bar"}


def test_enum():
    assert " " in str(CompareType.lower_threshold)
    assert " " in str(FlowType.dark_center)


class TestSettingsWidget:
    def test_create(self, qtbot):
        w = _settings.SettingsEditor()
        qtbot.addWidget(w.native)

    def test_change_units(self, qtbot):
        s = _settings.get_settings()
        s.io_units = Units.µm
        w = _settings.SettingsEditor()
        qtbot.addWidget(w.native)
        w.units_select.value = Units.nm
        assert s.io_units == Units.nm
        s.io_units = Units.mm
        assert w.units_select.value == Units.mm

    def test_change_layer_name_format(self, qtbot):
        s = _settings.get_settings()
        s.layer_naming_format = LayerNamingFormat.channel_only
        w = _settings.SettingsEditor()
        qtbot.addWidget(w.native)
        w.layer_naming_select.value = LayerNamingFormat.channel_filename
        assert s.layer_naming_format == LayerNamingFormat.channel_filename
        s.layer_naming_format = LayerNamingFormat.channel_only
        assert w.layer_naming_select.value == LayerNamingFormat.channel_only



import pytest

from PartSeg._roi_analysis.batch_window import CalculationPrepare
from PartSegCore.analysis.batch_processing.batch_backend import CalculationManager
from PartSegCore.analysis.calculation_plan import CalculationPlan, CalculationTree, MaskSuffix, RootType


@pytest.fixture
def calculation_prepare(tmp_path, part_settings, qtbot):
    def _constructor(file_list, calculation_plan=None):
        if calculation_plan is None:
            calculation_plan = CalculationPlan()
        dial = CalculationPrepare(
            file_list=file_list,
            calculation_plan=calculation_plan,
            measurement_file_path=tmp_path / "test1.xlsx",
            settings=part_settings,
            batch_manager=CalculationManager(),
        )
        qtbot.addWidget(dial)
        return dial

    return _constructor


class TestCalculationPrepare:
    @pytest.mark.parametrize("files_li", [["test1.tif", "test2.tif"], ["test1.tif"]])
    def test_init(self, calculation_prepare, tmp_path, files_li):
        files = [tmp_path / x for x in files_li]
        for file in files:
            file.write_text("test")
        dial = calculation_prepare(file_list=files)
        assert dial.all_file_prefix == str(tmp_path)
        assert dial.execute_btn.isEnabled()

    def test_show(self, calculation_prepare, tmp_path):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif", tmp_path / "test2.aaa"]
        for file in files:
            file.write_text("test")
        dial = calculation_prepare(file_list=files)
        assert dial.file_list_widget.topLevelItemCount() == 0
        dial._show_event_setup()
        assert dial.file_list_widget.topLevelItemCount() == 3

    def test_no_file(self, calculation_prepare, tmp_path):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif"]
        for file in files:
            file.write_text("test")

        files.append(tmp_path / "test3.tif")

        dial = calculation_prepare(file_list=files)
        assert not dial.execute_btn.isEnabled()

    def test_no_mask(self, calculation_prepare, tmp_path):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif"]
        for file in files:
            file.write_text("test")
            file.with_name(file.with_suffix("").name + "_mask.tif").write_text("test")

        files.append(tmp_path / "test3.tif")
        files[-1].write_text("test")

        dial = calculation_prepare(
            file_list=files,
            calculation_plan=CalculationPlan(
                CalculationTree(RootType.Image, [CalculationTree(MaskSuffix(name="", suffix="_mask"), [])])
            ),
        )
        dial._check_start_conditions()
        assert not dial.execute_btn.isEnabled()

    def test_choose_data_prefix(self, calculation_prepare, tmp_path, monkeypatch):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif"]
        for file in files:
            file.write_text("test")
        monkeypatch.setattr("PartSeg._roi_analysis.batch_window.QFileDialog.exec_", lambda *_: True)
        monkeypatch.setattr(
            "PartSeg._roi_analysis.batch_window.QFileDialog.selectedFiles", lambda *_: [tmp_path / "test_dir"]
        )
        dial = calculation_prepare(file_list=files)
        dial.choose_data_prefix()

        assert dial.base_prefix.text() == str(tmp_path / "test_dir")

    def test_choose_result_prefix(self, calculation_prepare, tmp_path, monkeypatch):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif"]
        for file in files:
            file.write_text("test")
        monkeypatch.setattr("PartSeg._roi_analysis.batch_window.QFileDialog.exec_", lambda *_: True)
        monkeypatch.setattr(
            "PartSeg._roi_analysis.batch_window.QFileDialog.selectedFiles", lambda *_: [tmp_path / "test_dir"]
        )
        dial = calculation_prepare(file_list=files)
        dial.choose_result_prefix()

        assert dial.result_prefix.text() == str(tmp_path / "test_dir")

    def test_overwrite_voxel_size(self, calculation_prepare, tmp_path):
        files = [tmp_path / "test1.tif", tmp_path / "test2.tif"]
        for file in files:
            file.write_text("test")
        dial = calculation_prepare(file_list=files)
        assert "ignored" not in dial.info_label.text()
        dial.overwrite_voxel_size_check.setChecked(True)
        assert "ignored" in dial.info_label.text()
        dial.overwrite_voxel_size_check.setChecked(False)
        assert "ignored" not in dial.info_label.text()



"""
This module contains non gui Qt based components
"""

import contextlib
import os.path
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from napari.settings import NapariSettings


def napari_get_settings(path=None) -> "NapariSettings":
    from napari.settings import get_settings as _napari_get_settings

    if path is not None:
        path = os.path.join(path, "settings.yaml")
    if path is not None:
        with contextlib.suppress(Exception):
            return _napari_get_settings(path)
    return _napari_get_settings()


__all__ = ("napari_get_settings",)



from typing import Callable, Optional, Union

from scipy.ndimage import zoom

from PartSegCore.algorithm_describe_base import AlgorithmProperty
from PartSegCore.image_transforming.transform_base import TransformBase
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


class InterpolateImage(TransformBase):
    @classmethod
    def get_fields(cls):
        return ["It can be very slow.", AlgorithmProperty("scale", "Scale", 1.0)]

    @classmethod
    def get_fields_per_dimension(cls, image: Image) -> list[Union[str, AlgorithmProperty]]:
        component_list = list(image.get_dimension_letters())
        return [
            "it can be very slow",
            *[AlgorithmProperty(f"scale_{i.lower()}", f"Scale {i}", 1.0) for i in reversed(component_list)],
        ]

    @classmethod
    def get_name(cls):
        return "Spline Interpolate"

    @classmethod
    def transform(
        cls,
        image: Image,
        roi_info: Optional[ROIInfo],
        arguments: dict,
        callback_function: Optional[Callable[[str, int], None]] = None,
    ) -> tuple[Image, Optional[ROIInfo]]:
        keys = [x for x in arguments if x.startswith("scale")]
        keys_order = Image.axis_order.lower()
        scale_factor = [1.0] * len(keys_order)
        if len(keys) == 1 and keys[0] == "scale":
            for letter in image.get_dimension_letters().lower():
                scale_factor[keys_order.index(letter)] = arguments["scale"]
            spacing = [x / arguments["scale"] for x in image.spacing]
        else:
            # assume that all keys are in format scale_{}
            for key in keys:
                letter = key[-1]
                scale_factor[keys_order.index(letter)] = arguments[key]
            spacing = [
                x / arguments[f"scale_{y}"] for x, y in zip(image.spacing, image.get_dimension_letters().lower())
            ]
        array = zoom(image.get_data(), scale_factor, mode="mirror")
        if image.mask is not None:
            mask = zoom(image.mask, scale_factor[:-1], mode="mirror")
        else:
            mask = None
        return image.substitute(data=array, image_spacing=spacing, mask=mask), None

    @classmethod
    def calculate_initial(cls, image: Image):
        min_val = min(image.spacing)
        return {
            f"scale_{letter}": x / min_val for x, letter in zip(image.spacing, image.get_dimension_letters().lower())
        }



import typing
from abc import ABC, abstractmethod
from io import BytesIO
from itertools import product
from pathlib import Path

import numpy as np
from tifffile import imwrite

from PartSegImage.image import Image, minimal_dtype


class BaseImageWriter(ABC):
    @classmethod
    @abstractmethod
    def save(cls, image: Image, save_path: typing.Union[str, BytesIO, Path], compression="ADOBE_DEFLATE"):
        pass

    @classmethod
    @abstractmethod
    def save_mask(cls, image: Image, save_path: typing.Union[str, Path], compression="ADOBE_DEFLATE"):
        pass


class ImageWriter(BaseImageWriter):
    """class for saving TIFF images"""

    @classmethod
    def prepare_metadata(cls, image: Image, channels: int):
        spacing = image.get_um_spacing()
        shift = image.get_um_shift()
        plane_li = [
            {
                "TheT": t,  # codespell:ignore thet
                "TheZ": z,
                "TheC": c,
                "PositionZ": shift[0] if len(shift) == 3 else 0,
                "PositionY": shift[-2],
                "PositionX": shift[-1],
                "PositionZUnit": "µm",
                "PositionYUnit": "µm",
                "PositionXUnit": "µm",
            }
            for t, z, c in product(range(image.times), range(image.layers), range(channels))
        ]

        metadata = {
            "Pixels": {
                "PhysicalSizeZ": spacing[0] if len(spacing) == 3 else 1,
                "PhysicalSizeY": spacing[-2],
                "PhysicalSizeX": spacing[-1],
                "PhysicalSizeZUnit": "µm",
                "PhysicalSizeYUnit": "µm",
                "PhysicalSizeXUnit": "µm",
            },
            "Plane": plane_li,
            "Creator": "PartSeg",
        }
        if image.name:
            metadata["Name"] = image.name

        return metadata

    @classmethod
    def save(cls, image: Image, save_path: typing.Union[str, BytesIO, Path], compression="ADOBE_DEFLATE"):
        """
        Save image as tiff to path or buffer

        :param image: image for save
        :param save_path: save location
        """
        data = image.get_image_for_save()

        metadata = cls.prepare_metadata(image, image.channels)

        metadata["Channel"] = {
            "Name": image.channel_names,
            "axes": "TZYXC",
        }
        cls._save(data, save_path, metadata, compression)

    @classmethod
    def save_mask(cls, image: Image, save_path: typing.Union[str, Path], compression="ADOBE_DEFLATE"):
        """
        Save mask connected to image as tiff to path or buffer

        :param image: mast is obtain with :py:meth:`.Image.get_mask_for_save`
        :param save_path: save location
        """
        mask = image.get_mask_for_save()
        if mask is None:
            return
        mask_max = np.max(mask)
        mask = mask.astype(minimal_dtype(mask_max))
        metadata = cls.prepare_metadata(image, 1)
        metadata["Channel"] = {
            "Name": "Mask",
            "axes": "TZYX",
        }
        cls._save(mask, save_path, metadata, compression)

    @staticmethod
    def _save(data: np.ndarray, save_path, metadata=None, compression="ADOBE_DEFLATE"):
        # TODO change to ome TIFF
        imwrite(
            save_path,
            data,
            ome=True,
            software="PartSeg",
            metadata=metadata,
            compression=compression,
        )  # , compress=6,


class IMAGEJImageWriter(BaseImageWriter):
    """class for saving TIFF images"""

    @classmethod
    def save(cls, image: Image, save_path: typing.Union[str, BytesIO, Path], compression=""):
        """
        Save image as tiff to path or buffer
        :param image: image for save
        :param save_path: save location
        """
        data = image.get_image_for_save()
        spacing = image.get_um_spacing()
        metadata: dict[str, typing.Any] = {"mode": "color", "unit": "\\u00B5m"}
        if len(spacing) == 3:
            metadata["spacing"] = spacing[0]
        if image.channel_names is not None:
            metadata["Labels"] = image.channel_names * image.layers
        coloring = image.get_imagej_colors()
        if coloring is not None:
            metadata["LUTs"] = coloring
        ranges_li = image.get_ranges()
        metadata["Ranges"] = np.array(ranges_li).reshape(len(ranges_li) * 2)

        resolution = [1 / x for x in spacing[-2:]]
        cls._save(data, save_path, resolution, metadata)

    @classmethod
    def save_mask(cls, image: Image, save_path: typing.Union[str, Path], compression=""):
        """
        Save mask connected to image as tiff to path or buffer
        :param image: mast is obtain with :py:meth:`.Image.get_mask_for_save`
        :param save_path: save location
        """
        mask = image.get_mask_for_save()
        if mask is None:
            return
        mask_max = np.max(mask)
        mask = mask.astype(minimal_dtype(mask_max))
        metadata: dict[str, typing.Union[str, float]] = {"mode": "color", "unit": "\\u00B5m"}
        spacing = image.get_um_spacing()
        if len(spacing) == 3:
            metadata["spacing"] = spacing[0]
        resolution = [1 / x for x in spacing[-2:]]
        cls._save(mask, save_path, resolution, metadata)

    @staticmethod
    def _save(data: np.ndarray, save_path, resolution=None, metadata=None):
        # TODO change to ome TIFF
        if data.dtype in [np.uint8, np.uint16, np.float32]:
            imwrite(
                save_path,
                data,
                imagej=True,
                software="PartSeg",
                metadata=metadata,
                resolution=resolution,
            )  # , compress=6,
        else:
            raise ValueError(f"Data type {data.dtype} not supported by imagej tiff")
            # imagej=True, software="PartSeg")



import json
import logging
import os
import os.path
import re
import sys
import warnings
from argparse import Namespace
from collections.abc import Sequence
from contextlib import suppress
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, ClassVar, NamedTuple, Optional, Union

import napari.utils.theme
import numpy as np
import pint
from napari.qt import get_stylesheet
from napari.utils import Colormap
from napari.utils.theme import get_theme
from napari.utils.theme import template as napari_template
from packaging.version import parse as parse_version
from qtpy.QtCore import QObject, Signal
from qtpy.QtWidgets import QMessageBox, QWidget

from PartSeg.common_backend import napari_get_settings
from PartSeg.common_backend.partially_const_dict import PartiallyConstDict
from PartSegCore import register
from PartSegCore.color_image import default_colormap_dict, default_label_dict
from PartSegCore.color_image.base_colors import starting_colors
from PartSegCore.io_utils import find_problematic_entries, load_metadata_base
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.project_info import AdditionalLayerDescription, HistoryElement, ProjectInfoBase
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult
from PartSegCore.utils import ProfileDict
from PartSegImage import Image

if TYPE_CHECKING:  # pragma: no cover
    from napari.settings import NapariSettings
logger = logging.getLogger(__name__)

_napari_ge_5 = parse_version(napari.__version__) >= parse_version("0.5.0a1")

DIR_HISTORY = "io.dir_location_history"
FILE_HISTORY = "io.files_open_history"
MULTIPLE_FILES_OPEN_HISTORY = "io.multiple_files_open_history"
ROI_NOT_FIT = "roi do not fit to image"
IO_SAVE_DIRECTORY = "io.save_directory"


class ImageSettings(QObject):
    """
    Base class for all PartSeg settings. Keeps information about current Image.
    """

    image_changed = Signal(Image)
    image_path_changed = Signal(str)
    image_channel_count_changed = Signal(int)
    image_spacing_changed = Signal()
    roi_changed = Signal(ROIInfo)
    """
    :py:class:`.Signal`
    emitted when roi has changed
    """
    roi_clean = Signal()
    additional_layers_changed = Signal()

    def __init__(self):
        super().__init__()
        self._image: Optional[Image] = None
        self._image_path = ""
        self._roi_info = ROIInfo(None)
        self._additional_layers = {}
        self._parent: Optional[QWidget] = None

    def set_parent(self, parent: QWidget):
        self._parent = parent

    @property
    def full_segmentation(self):  # pragma: no cover
        raise AttributeError("full_segmentation not supported")

    @full_segmentation.setter
    def full_segmentation(self, val):  # pragma: no cover # pylint: disable=no-self-use
        raise AttributeError("full_segmentation not supported")

    @property
    def noise_remove_image_part(self):  # pragma: no cover
        raise AttributeError("noise_remove_image_part not supported")

    @noise_remove_image_part.setter
    def noise_remove_image_part(self, val):  # pragma: no cover # pylint: disable=no-self-use
        raise AttributeError("noise_remove_image_part not supported")

    @property
    def additional_layers(self) -> dict[str, AdditionalLayerDescription]:
        return self._additional_layers

    @additional_layers.setter
    def additional_layers(self, val):  # pragma: no cover  # pylint: disable=no-self-use
        raise AttributeError("additional_layers assign not supported")

    @property
    def image_spacing(self):
        """:py:meth:`Image.spacing` proxy"""
        return self._image.spacing if self._image is not None else ()

    def is_image_2d(self):
        """:py:meth:`Image.is_2d` proxy"""
        return self._image is None or self._image.is_2d

    @image_spacing.setter
    def image_spacing(self, value):
        if len(value) not in [2, 3]:
            raise ValueError(f"value parameter should have length 2 or 3. Current length is {len(value)}.")
        if len(value) == 2:
            self._image.set_spacing((self._image.spacing[0], *list(value)))
        else:
            self._image.set_spacing(value)
        self.image_spacing_changed.emit()

    @property
    def segmentation(self) -> np.ndarray:  # pragma: no cover
        """current roi"""
        warnings.warn("segmentation parameter is renamed to roi", DeprecationWarning, stacklevel=2)
        return self.roi

    @property
    def roi(self) -> np.ndarray:
        """current roi"""
        return self._roi_info.roi

    @property
    def segmentation_info(self) -> ROIInfo:  # pragma: no cover
        warnings.warn("segmentation info parameter is renamed to roi", DeprecationWarning, stacklevel=2)
        return self.roi_info

    @property
    def roi_info(self) -> ROIInfo:
        return self._roi_info

    @roi.setter
    def roi(self, val: Union[np.ndarray, ROIInfo]):
        if val is None:
            self._roi_info = ROIInfo(val)
            self._additional_layers = {}
            self.roi_clean.emit()
            return
        try:
            if isinstance(val, np.ndarray):
                self._roi_info = ROIInfo(self.image.fit_array_to_image(val))
            else:
                self._roi_info = val.fit_to_image(self.image)
        except ValueError as e:
            raise ValueError(ROI_NOT_FIT) from e
        self._additional_layers = {}
        self.roi_changed.emit(self._roi_info)

    @property
    def sizes(self):
        return self._roi_info.sizes

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value: Image):
        if value is None:
            return
        self._image = value
        if value.file_path is not None:
            self.image_path_changed.emit(value.file_path)
        self._image_changed()
        self._roi_info = ROIInfo(None)

        self.image_changed.emit(self._image)
        self.image_channel_count_changed.emit(self._image.channels)

    @property
    def has_channels(self):
        return self.channels > 1

    def _image_changed(self):
        """Reimplement hook for change of main image"""

    @property
    def image_path(self):
        return self._image.file_path if self.image is not None else ""

    @property
    def image_shape(self):
        # TODO analyse and decide if channels should be part of shape
        return self._image.shape if self.image is not None else ()

    @image_path.setter
    def image_path(self, value):
        self._image.file_path = value
        self.image_path_changed.emit(self._image_path)

    @property
    def channels(self):
        return 0 if self._image is None else self._image.channels

    def components_mask(self):
        return np.array([0] + [1] * np.max(self.roi), dtype=np.uint8)


class ColormapDict(PartiallyConstDict[Colormap]):
    """
    Dict for mixing custom colormap with predefined ones
    """

    if os.path.basename(sys.argv[0]) in ["sphinx-build", "sphinx-build.exe"]:  # pragma: no cover
        const_item_dict = {}
    else:
        const_item_dict = default_colormap_dict
    """
    Non removable items for this dict. Current value is :py:data:`default_colormap_dict`
    """

    @property
    def colormap_removed(self):
        """
        Signal that colormap is removed form dict
        """
        return self.item_removed

    @property
    def colormap_added(self):
        """
        Signal that colormap is added to dict
        """
        return self.item_added


class LabelColorDict(PartiallyConstDict[list]):
    """
    Dict for mixing custom label colors with predefined ones`
    """

    const_item_dict = default_label_dict
    """Non removable items for this dict. Current value is :py:data:`default_label_dict`"""

    def get_array(self, key: str) -> np.ndarray:
        """Get labels as numpy array"""
        return np.array(self[key][0], dtype=np.uint8)


class ViewSettings(ImageSettings):
    colormap_changes = Signal()
    labels_changed = Signal()
    theme_changed = Signal()
    profile_data_changed = Signal(str, object)
    """Signal about changes in stored data (set with set_in_profile)"""

    def __init__(self):
        super().__init__()
        self.color_map = []
        self.border_val = []
        self.current_profile_dict = "default"
        self.view_settings_dict = ProfileDict()
        self.colormap_dict = ColormapDict(self.get_from_profile("custom_colormap", {}))
        self.label_color_dict = LabelColorDict(self.get_from_profile("custom_label_colors", {}))
        self.cached_labels: Optional[tuple[str, np.ndarray]] = None

    @property
    def theme_name(self) -> str:
        """Name of current theme."""
        return self.get_from_profile("theme", "dark")

    @property
    def theme(self):
        """Theme as structure."""
        if _napari_ge_5:
            return get_theme(self.theme_name)
        try:
            return get_theme(self.theme_name, as_dict=False)
        except TypeError:  # pragma: no cover
            theme = get_theme(self.theme_name)
            return Namespace(
                **{k: Color(v) if isinstance(v, str) and v.startswith("rgb") else v for k, v in theme.items()}
            )

    @property
    def style_sheet(self):
        return self.get_style_sheet()

    def get_style_sheet(self):
        """QSS style sheet for current theme."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)
            if _napari_ge_5:
                theme = get_theme(self.theme_name).to_rgb_dict()
            else:
                theme = get_theme(self.theme_name, as_dict=True)
        # TODO understand qss overwrite mechanism
        return napari_template("\n".join(register.qss_list) + get_stylesheet() + "\n".join(register.qss_list), **theme)

    @theme_name.setter
    def theme_name(self, value: str):
        """Name of current theme."""
        if value not in napari.utils.theme.available_themes():
            raise ValueError(f"Unsupported theme {value}. Supported one: {self.theme_list()}")
        if value == self.theme_name:
            return
        self.set_in_profile("theme", value)
        self.theme_changed.emit()

    @staticmethod
    def theme_list():
        """Sequence of available themes"""
        try:
            return napari.utils.theme.available_themes()
        except:  # noqa: E722  # pylint: disable=bare-except  # pragma: no cover
            return ("light",)

    @property
    def chosen_colormap(self):
        """Sequence of selected colormap to be available in dropdown"""
        data = self.get_from_profile("colormaps", starting_colors[:])
        res = [x for x in data if x in self.colormap_dict]
        if len(res) != data:
            if not res:
                res = starting_colors[:]
            self.set_in_profile("colormaps", res)
        return res

    @chosen_colormap.setter
    def chosen_colormap(self, val):
        self.set_in_profile("colormaps", val)
        self.colormap_changes.emit()

    @property
    def current_labels(self):
        """Current labels scheme for marking ROI"""
        return self.get_from_profile("labels_used", "default")

    @current_labels.setter
    def current_labels(self, val):
        if val not in self.label_color_dict:
            raise ValueError(f"Unknown label scheme name '{val}'")
        self.set_in_profile("labels_used", val)
        self.labels_changed.emit()

    @property
    def label_colors(self):
        key = self.current_labels
        if key not in self.label_color_dict:
            key = "default"
            self.current_labels = key

        if not (self.cached_labels and key == self.cached_labels[0]):
            self.cached_labels = key, self.label_color_dict.get_array(key)

        return self.cached_labels[1]

    def chosen_colormap_change(self, name, visibility):
        colormaps = set(self.chosen_colormap)
        if visibility:
            colormaps.add(name)
        else:
            with suppress(KeyError):
                colormaps.remove(name)
        # TODO update sorting rule
        self.chosen_colormap = sorted(colormaps, key=self.colormap_dict.get_position)

    def get_channel_info(self, view: str, num: int, default: Optional[str] = None) -> str:  # pragma: no cover
        warnings.warn(
            "get_channel_info is deprecated, use get_channel_colormap_name instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.get_channel_colormap_name(view, num, default)

    def get_channel_colormap_name(self, view: str, num: int, default: Optional[str] = None) -> str:
        cm = self.chosen_colormap
        if default is None:
            default = cm[num % len(cm)]
        resp = self.get_from_profile(f"{view}.cmap.num{num}", default)
        if resp not in self.colormap_dict:
            resp = cm[num % len(cm)]
            self.set_in_profile(f"{view}.cmap.num{num}", resp)
        return resp

    def set_channel_info(self, view: str, num, value: str):  # pragma: no cover
        warnings.warn(
            "set_channel_info is deprecated, use set_channel_colormap_name instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.set_channel_colormap_name(view, num, value)

    def set_channel_colormap_name(self, view: str, num, value: str):
        self.set_in_profile(f"{view}.cmap.num{num}", value)

    def connect_channel_colormap_name(self, view: str, fun: Callable):
        self.connect_to_profile(f"{view}.cmap", fun)

    @property
    def available_colormaps(self):
        return list(self.colormap_dict.keys())

    def _image_changed(self):
        self.border_val = self.image.get_ranges()
        super()._image_changed()

    def change_profile(self, name):
        self.current_profile_dict = name
        self.view_settings_dict.profile_change()

    def set_in_profile(self, key_path, value):
        """
        Function for saving information used in visualization. This is accessor to
        :py:meth:`~.ProfileDict.set` of inner variable.

        :param key_path: dot separated path
        :param value: value to store. The value need to be json serializable."""
        self.view_settings_dict.set(f"{self.current_profile_dict}.{key_path}", value)
        self.profile_data_changed.emit(key_path, value)

    def get_from_profile(self, key_path, default=None):
        """
        Function for getting information used in visualization. This is accessor to
        :py:meth:`~.ProfileDict.get` of inner variable.

        :param key_path: dot separated path
        :param default: default value if key is missed
        """
        return self.view_settings_dict.get(f"{self.current_profile_dict}.{key_path}", default)

    def connect_to_profile(self, key_path, callback):
        # TODO  fixme fix when introduce switch profiles
        self.view_settings_dict.connect(key_path, callback)


class SaveSettingsDescription(NamedTuple):
    file_name: str
    values: Union[dict, ProfileDict]


class BaseSettings(ViewSettings):
    """
    :ivar json_folder_path: default location for saving/loading settings data
    :ivar last_executed_algorithm: name of last executed algorithm.
    :cvar save_locations_keys: list of names of distinct save location.
        location are stored in "io"

    """

    mask_changed = Signal()
    points_changed = Signal()
    request_load_files = Signal(list)
    """:py:class:`~.Signal` mask changed signal"""
    json_encoder_class = PartSegEncoder
    load_metadata = staticmethod(load_metadata_base)
    algorithm_changed = Signal()
    """:py:class:`~.Signal` emitted when current algorithm should be changed"""
    save_locations_keys: ClassVar[list[str]] = []

    def __init__(self, json_path: Union[Path, str], profile_name: str = "default"):
        """
        :param json_path: path to store
        :param profile_name: name of profile to be used. default value is "default"
        """
        super().__init__()
        napari_path = os.path.dirname(json_path) if os.path.basename(json_path) in ["analysis", "mask"] else json_path
        self.napari_settings: NapariSettings = napari_get_settings(napari_path)
        self._current_roi_dict = profile_name
        self._roi_dict = ProfileDict()
        self._last_algorithm_dict = ProfileDict()
        self.json_folder_path = json_path
        self.last_executed_algorithm = ""
        self.history: list[HistoryElement] = []
        self.history_index = -1
        self.last_executed_algorithm = ""
        self._points = None
        pint.get_application_registry()("nm")  # enforce pint registry initialization

    def _image_changed(self):
        super()._image_changed()
        self.points = None

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, value):
        self._points = value if value is not None else None
        self.points_changed.emit()

    @property
    def theme_name(self) -> str:
        try:
            theme = self.napari_settings.appearance.theme
            if self.get_from_profile("first_start", True):
                theme = "light"
                self.napari_settings.appearance.theme = theme
                self.set_in_profile("first_start", False)
            return theme
        except AttributeError:  # pragma: no cover
            return "light"

    @theme_name.setter
    def theme_name(self, value: str):
        self.napari_settings.appearance.theme = value

    def set_segmentation_result(self, result: ROIExtractionResult):
        if (
            result.file_path is not None and result.file_path and result.file_path != self.image.file_path
        ):  # pragma: no cover
            if self._parent is not None:
                # TODO change to non disrupting popup
                QMessageBox().warning(
                    self._parent, "Result file bug", "It looks like one try to set ROI form another file."
                )
            return
        if result.info_text and self._parent is not None:
            QMessageBox().information(self._parent, "Algorithm info", result.info_text)

        self._additional_layers = result.additional_layers
        self.last_executed_algorithm = result.parameters.algorithm
        self.set_algorithm(f"algorithms.{result.parameters.algorithm}", result.parameters.values)
        # Fixme not use EventedDict here
        try:
            roi_info = result.roi_info.fit_to_image(self.image)
        except ValueError as e:  # pragma: no cover
            raise ValueError(ROI_NOT_FIT) from e
        if result.points is not None:
            self.points = result.points
        self._roi_info = roi_info
        self.roi_changed.emit(self._roi_info)

    def _load_files_call(self, files_list: list[str]):
        self.request_load_files.emit(files_list)

    def add_history_element(self, elem: HistoryElement) -> None:
        self.history_index += 1
        if self.history_index < len(self.history) and self.cmp_history_element(elem, self.history[self.history_index]):
            self.history[self.history_index] = elem
        else:
            self.history = self.history[: self.history_index]
            self.history.append(elem)

    def history_size(self) -> int:
        return self.history_index + 1

    def history_redo_size(self) -> int:
        if self.history_index + 1 == len(self.history):
            return 0
        return len(self.history[self.history_index + 1 :])

    def history_redo_clean(self) -> None:
        self.history = self.history[: self.history_size()]

    def history_current_element(self) -> HistoryElement:
        return self.history[self.history_index]

    def history_next_element(self) -> HistoryElement:
        return self.history[self.history_index + 1]

    def history_pop(self) -> Optional[HistoryElement]:
        if self.history_index != -1:
            self.history_index -= 1
            return self.history[self.history_index + 1]
        return None

    def set_history(self, history: list[HistoryElement]):
        self.history = history
        self.history_index = len(self.history) - 1

    def get_history(self) -> list[HistoryElement]:
        return self.history[: self.history_index + 1]

    @staticmethod
    def cmp_history_element(el1, el2):
        return False

    @property
    def mask(self):
        return self._image.mask

    @mask.setter
    def mask(self, value):
        try:
            self._image.set_mask(value)
            self.mask_changed.emit()
        except ValueError as e:
            raise ValueError("mask do not fit to image") from e

    def get_save_list(self) -> list[SaveSettingsDescription]:
        """List of files in which program save the state."""
        return [
            SaveSettingsDescription("segmentation_settings.json", self._roi_dict),
            SaveSettingsDescription("view_settings.json", self.view_settings_dict),
            SaveSettingsDescription("algorithm_settings.json", self._last_algorithm_dict),
        ]

    def get_path_history(self) -> list[str]:
        """
        return list containing last 10 elements added with :py:meth:`.add_path_history` and
        last opened in each category form :py:attr:`save_location_keys`
        """
        res = self.get(DIR_HISTORY, [])[:]
        for name in self.save_locations_keys:
            val = self.get(f"io.{name}", str(Path.home()))
            if val not in res:
                res = [*res, val]
        return res

    @staticmethod
    def _add_elem_to_list(data_list: list, value: Any, keep_len=10) -> list:
        try:
            data_list.remove(value)
        except ValueError:
            data_list = data_list[: keep_len - 1]
        return [value, *data_list]

    def get_last_files(self) -> list[tuple[tuple[Union[str, Path], ...], str]]:
        return self.get(FILE_HISTORY, [])

    def add_load_files_history(self, file_path: Sequence[Union[str, Path]], load_method: str):  # pragma: no cover
        warnings.warn("`add_load_files_history` is deprecated", FutureWarning, stacklevel=2)
        return self.add_last_files(file_path, load_method)

    def add_last_files(self, file_path: Sequence[Union[str, Path]], load_method: str):
        self.set(FILE_HISTORY, self._add_elem_to_list(self.get(FILE_HISTORY, []), [list(file_path), load_method]))
        # keep list of files as list because json serialize tuple to list
        self.add_path_history(os.path.dirname(file_path[0]))

    def get_last_files_multiple(self) -> list[tuple[tuple[Union[str, Path], ...], str]]:
        return self.get(MULTIPLE_FILES_OPEN_HISTORY, [])

    def add_last_files_multiple(self, file_paths: list[Union[str, Path]], load_method: str):
        self.set(
            MULTIPLE_FILES_OPEN_HISTORY,
            self._add_elem_to_list(
                self.get(MULTIPLE_FILES_OPEN_HISTORY, []), [list(file_paths), load_method], keep_len=30
            ),
        )
        # keep list of files as list because json serialize tuple to list
        self.add_path_history(os.path.dirname(file_paths[0]))

    def add_path_history(self, dir_path: Union[str, Path]):
        """Save path in history of visited directories. Store only 10 last"""
        dir_path = str(dir_path)
        self.set(DIR_HISTORY, self._add_elem_to_list(self.get(DIR_HISTORY, []), dir_path))

    def set(self, key_path: str, value):
        """
        function for saving general state (not visualization). This is accessor to
        :py:meth:`~.ProfileDict.set` of inner variable.

        :param key_path: dot separated path
        :param value: value to store. The value need to be json serializable.
        """
        if key_path.startswith(("algorithm_widget_state.", "algorithms.")) or key_path == "current_algorithm":
            warnings.warn("Use `set_algorithm_state` instead of `set` for algorithm state", FutureWarning, stacklevel=2)
            self.set_algorithm(key_path, value)
            return
        self._roi_dict.set(f"{self._current_roi_dict}.{key_path}", value)

    def get(self, key_path: str, default=None):
        """
        Function for getting general state (not visualization). This is accessor to
        :py:meth:`~.ProfileDict.get` of inner variable.

        :param key_path: dot separated path
        :param default: default value if key is missed
        """
        if key_path.startswith(("algorithms.", "algorithm_widget_state.")) or key_path == "current_algorithm":
            warnings.warn("Use `set_algorithm_state` instead of `set` for algorithm state", FutureWarning, stacklevel=2)
            return self.get_algorithm(key_path, default)
        return self._roi_dict.get(f"{self._current_roi_dict}.{key_path}", default)

    def set_algorithm(self, key_path: str, value):
        """
        function for saving last algorithm used information. This is accessor to
        :py:meth:`~.ProfileDict.set` of inner variable.

        :param key_path: dot separated path
        :param value: value to store. The value need to be json serializable.
        """
        # if key_path.startswith("")
        self._last_algorithm_dict.set(f"{self._current_roi_dict}.{key_path}", value)

    def get_algorithm(self, key_path: str, default=None):
        """
        Function for getting last algorithm used information. This is accessor to
        :py:meth:`~.ProfileDict.get` of inner variable.

        :param key_path: dot separated path
        :param default: default value if key is missed
        """
        return self._last_algorithm_dict.get(f"{self._current_roi_dict}.{key_path}", default)

    def connect_(self, key_path, callback):
        # TODO  fixme fix when introduce switch profiles
        self._roi_dict.connect(key_path, callback)

    def dump_part(self, file_path, path_in_dict, names=None):
        data = self.get(path_in_dict)
        if names is not None:
            data = {name: data[name] for name in names}
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as ff:
            json.dump(data, ff, cls=self.json_encoder_class, indent=2)

    def dump(self, folder_path: Union[Path, str, None] = None):
        """
        Save current application settings to disc.

        :param folder_path: path to directory in which data should be saved.
            If is None then use :py:attr:`.json_folder_path`
        """
        if self.napari_settings.save is not None:
            self.napari_settings.save()
        else:
            self.napari_settings._save()  # pylint: disable=protected-access
        if folder_path is None:
            folder_path = self.json_folder_path
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        errors_list = []
        for el in self.get_save_list():
            try:
                dump_string = json.dumps(el.values, cls=self.json_encoder_class, indent=2)
                with open(os.path.join(folder_path, el.file_name), "w", encoding="utf-8") as ff:
                    ff.write(dump_string)
            except Exception as e:  # pylint: disable=broad-except   # noqa: PERF203
                errors_list.append((e, os.path.join(folder_path, el.file_name)))
        if errors_list:
            logger.error(errors_list)
        return errors_list

    def _load_settings_file(self, file_path: Union[Path, str]) -> tuple[ProfileDict, Any]:
        error = None
        data: ProfileDict = self.load_metadata(file_path)
        if isinstance(data, dict) and "__error__" in data:
            error = data["__error__"]
            data = ProfileDict()
        elif not data.verify_data():
            filtered = data.pop_errors()
            error = filtered
            filtered_base_str = ((k, "\n".join(f"{x}" for x in find_problematic_entries(v))) for k, v in filtered)
            filtered_str = "\n".join(f"{k}\n{v}\n" for k, v in filtered_base_str)

            logger.error("error in load data from %s problematic keys are %s", file_path, filtered_str)
        return data, error

    def load(self, folder_path: Union[Path, str, None] = None):
        """
        Load settings state from given directory

        :param folder_path: path to directory in which data should be saved.
            If is None then use :py:attr:`.json_folder_path`
        """
        if folder_path is None:
            folder_path = self.json_folder_path
        errors_dict = {}
        for el in self.get_save_list():
            file_path = os.path.join(folder_path, el.file_name)
            if not os.path.exists(file_path):
                continue
            error = None
            try:
                data, error = self._load_settings_file(file_path)
                if error is not None:
                    errors_dict[file_path] = error
                el.values.update(data)
            except Exception as e:  # pylint: disable=broad-except
                error = True
                logger.error(e)
                errors_dict[file_path] = e
            finally:
                if error is not None:
                    timestamp = datetime.now().strftime("%Y-%m-%d_%H_%M_%S")
                    base_path, ext = os.path.splitext(file_path)
                    os.rename(file_path, f"{base_path}_{timestamp}{ext}")

        self.label_color_dict._refresh_order()  # pylint: disable=protected-access
        self.colormap_dict._refresh_order()  # pylint: disable=protected-access

        return errors_dict

    def get_project_info(self) -> ProjectInfoBase:
        """Get all information needed to save project"""
        raise NotImplementedError  # pragma: no cover

    def set_project_info(self, data: ProjectInfoBase):
        """Set project info"""
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def verify_image(image: Image, silent=True) -> Union[Image, bool]:
        if image.is_time:
            if image.is_stack:
                raise TimeAndStackException
            if silent:
                return image.swap_time_and_stack()
            raise SwapTimeStackException
        return True


class SwapTimeStackException(Exception):
    """
    Exception which inform that current image shape is not supported,
    but can be if time and stack axes were swapped
    """


class TimeAndStackException(Exception):
    """
    Exception which inform that current image has both time
    and stack dat which is not supported
    """


class Color:  # pragma: no cover
    def __init__(self, text):
        color_re = re.compile(r"rgb\((\d+), (\d+), (\d+)\)")
        self.colors = tuple(map(int, color_re.match(text).groups()))

    def as_rgb_tuple(self):
        return self.colors



# pylint: disable=no-self-use

import os
from copy import deepcopy

import numpy as np
import pytest

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import (
    AnalysisAlgorithmSelection,
    ProjectTuple,
    SegmentationPipeline,
    SegmentationPipelineElement,
)
from PartSegCore.analysis.calculate_pipeline import calculate_pipeline
from PartSegCore.analysis.load_functions import LoadProject
from PartSegCore.analysis.save_functions import SaveProject
from PartSegCore.mask_create import calculate_mask


@pytest.mark.parametrize("channel", [0, 1])
def test_simple(image, algorithm_parameters, channel):
    algorithm_parameters["values"]["channel"] = channel
    algorithm = AnalysisAlgorithmSelection[algorithm_parameters["algorithm_name"]]()
    algorithm.set_image(image)
    algorithm.set_parameters(**algorithm_parameters["values"])
    result = algorithm.calculation_run(lambda x, y: None)
    assert np.max(result.roi) == 1


def test_pipeline_manual(image, algorithm_parameters, mask_property):
    algorithm = AnalysisAlgorithmSelection[algorithm_parameters["algorithm_name"]]()
    algorithm.set_image(image)
    algorithm.set_parameters(**algorithm_parameters["values"])
    result = algorithm.calculation_run(lambda x, y: None)
    mask = calculate_mask(mask_property, result.roi, None, image.spacing)
    algorithm_parameters["values"]["channel"] = 1
    algorithm.set_parameters(**algorithm_parameters["values"])
    algorithm.set_mask(mask)
    result2 = algorithm.calculation_run(lambda x, y: None)
    assert np.max(result2.roi) == 2


@pytest.mark.parametrize("use_mask", [True, False])
def test_pipeline(image, algorithm_parameters, mask_property, tmp_path, use_mask):
    elem = SegmentationPipelineElement(
        segmentation=ROIExtractionProfile(
            name="", algorithm=algorithm_parameters["algorithm_name"], values=algorithm_parameters["values"]
        ),
        mask_property=mask_property,
    )
    algorithm_parameters = deepcopy(algorithm_parameters)
    algorithm_parameters["values"]["channel"] = 1
    pipeline = SegmentationPipeline(
        name="",
        segmentation=ROIExtractionProfile(
            name="", algorithm=algorithm_parameters["algorithm_name"], values=algorithm_parameters["values"]
        ),
        mask_history=[elem],
    )
    mask = np.ones(image.get_channel(0).shape, dtype=np.uint8) if use_mask else None
    result = calculate_pipeline(image, mask, pipeline, lambda x, y: None)
    assert np.max(result.roi_info.roi) == 2
    pt = ProjectTuple(
        file_path=image.file_path,
        image=image,
        roi_info=result.roi_info,
        mask=result.mask,
        history=result.history,
        algorithm_parameters=algorithm_parameters,
    )
    SaveProject.save(tmp_path / "project.tgz", pt)
    assert os.path.exists(tmp_path / "project.tgz")
    loaded = LoadProject.load([tmp_path / "project.tgz"])
    assert np.all(loaded.roi_info.roi == result.roi_info.roi)



import inspect
import math
import textwrap
import typing
import warnings
from abc import ABC, ABCMeta, abstractmethod
from collections.abc import MutableMapping
from functools import wraps
from importlib.metadata import version
from typing import Annotated

from local_migrator import REGISTER, class_to_str
from packaging.version import parse as parse_version
from pydantic import BaseModel as PydanticBaseModel
from pydantic import create_model, validator
from pydantic.fields import Field, FieldInfo

from PartSegCore.utils import BaseModel
from PartSegImage import Channel

if typing.TYPE_CHECKING:
    from pydantic.fields import ModelField

try:
    # pydantic 1
    from pydantic.fields import UndefinedType
    from pydantic.main import ModelMetaclass

    def field_serializer(_):
        def decorator(func):
            return func

        return decorator

except ImportError:
    # pydantic 2
    from pydantic import field_serializer
    from pydantic._internal._model_construction import ModelMetaclass
    from pydantic_core import PydanticUndefinedType as UndefinedType


class AlgorithmDescribeNotFound(Exception):
    """
    When algorithm description not found
    """


class AlgorithmProperty:
    """
    This class is used to verbose describe algorithm parameters

    :param str name: name of parameter used in code
    :param str user_name: name presented to user in interface
    :param default_value: initial value which be used during interface generation
    :param str help_text: toll tip presented to user when keep mouse over widget
    :type value_type: type
    """

    def __init__(
        self,
        name: str,
        user_name: str,
        default_value: typing.Union[str, float, object],
        options_range=None,
        possible_values=None,
        value_type=None,
        help_text="",
        per_dimension=False,
        mgi_options=None,
        **kwargs,
    ):
        if "property_type" in kwargs:
            warnings.warn("property_type is deprecated, use value_type instead", DeprecationWarning, stacklevel=2)
            value_type = kwargs["property_type"]
            del kwargs["property_type"]
        if kwargs:
            raise ValueError(", ".join(kwargs.keys()) + " are not expected")

        self.name = name
        self.user_name = user_name
        if isinstance(possible_values, list):
            self.value_type = list
        elif value_type is not None:
            self.value_type = value_type
        else:
            self.value_type = type(default_value)
        self.default_value = default_value
        self.range = tuple(self.value_type(x) for x in options_range) if options_range is not None else None
        self.possible_values = possible_values
        self.help_text = help_text
        self.per_dimension = per_dimension
        self.mgi_options = mgi_options if mgi_options is not None else {}
        if self.value_type is list and default_value not in possible_values:
            raise ValueError(f"default_value ({default_value}) should be one of possible values ({possible_values}).")

    def __repr__(self):
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}(name='{self.name}',"
            f" user_name='{self.user_name}', "
            f"default_value={self.default_value}, type={self.value_type}, range={self.range},"
            f"possible_values={self.possible_values})"
        )


class _GetDescriptionClass:
    __slots__ = ("_name",)

    def __init__(self):
        self._name = None

    def __set_name__(self, owner, name: str):
        if self._name is None:
            self._name = name

    def __get__(self, obj, klass):
        if klass is None:
            klass = type(obj)

        name = typing.cast("str", self._name)
        fields_dkt = {
            field.name: (
                Annotated[field.value_type, field.user_name, field.range, field.help_text],
                field.default_value,
            )
            for field in klass.get_fields()
            if not isinstance(field, str)
        }

        model = create_model(name, **fields_dkt)
        model.__qualname__ = f"{klass.__qualname__}.{name}"
        setattr(klass, name, model)
        return model


def _partial_abstractmethod(funcobj):
    funcobj.__is_partial_abstractmethod__ = True
    return funcobj


class AlgorithmDescribeBaseMeta(ABCMeta):
    def __new__(cls, name, bases, attrs, **kwargs):
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)
        if (
            not inspect.isabstract(cls2)
            and hasattr(cls2.get_fields, "__is_partial_abstractmethod__")
            and cls2.__argument_class__ is None
        ):
            raise RuntimeError("class need to have __argument_class__ set or get_fields functions defined")
        cls2.__new_style__ = getattr(cls2.get_fields, "__is_partial_abstractmethod__", False)
        return cls2


class AlgorithmDescribeBase(ABC, metaclass=AlgorithmDescribeBaseMeta):
    """
    This is abstract class for all algorithm exported to user interface.
    Based on get_name and get_fields methods the interface will be generated
    For each group of algorithm base abstract class will add additional methods
    """

    __argument_class__: typing.Optional[type[PydanticBaseModel]] = None
    __new_style__: bool

    @classmethod
    def get_doc_from_fields(cls):
        resp = "{\n"
        for el in get_fields_from_algorithm(cls):
            if isinstance(el, AlgorithmProperty):
                resp += f"  {el.name}: {el.value_type} - "
                if el.help_text:
                    resp += el.help_text
                resp += f"(default values: {el.default_value})\n"
        resp += "}\n"
        return resp

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        Algorithm name. It will be used during interface generating and in registering
        to proper :py:class:`PartSeg.PartSegCore.algorithm_describe_base.Register`.

        :return: name of algorithm
        """
        raise NotImplementedError

    @classmethod
    @_partial_abstractmethod
    def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
        """
        This function return list of parameters needed by algorithm. It is used for generate form in User Interface

        :return: list of algorithm parameters and comments
        """
        if hasattr(cls, "__argument_class__") and cls.__argument_class__ is not None:
            warnings.warn(
                "Class has __argument_class__ defined, one should not use get_fields",
                category=FutureWarning,
                stacklevel=2,
            )
            return base_model_to_algorithm_property(cls.__argument_class__)
        raise NotImplementedError

    @classmethod
    def _get_fields(cls):
        return base_model_to_algorithm_property(cls.__argument_class__) if cls.__new_style__ else cls.get_fields()

    @classmethod
    def get_fields_dict(cls) -> dict[str, AlgorithmProperty]:
        return {v.name: v for v in get_fields_from_algorithm(cls) if isinstance(v, AlgorithmProperty)}

    @classmethod
    def get_default_values(cls):
        if cls.__new_style__:
            return cls.__argument_class__()  # pylint: disable=not-callable
        return {
            el.name: (
                {
                    "name": el.default_value,
                    "values": el.possible_values[el.default_value].get_default_values(),
                }
                if issubclass(el.value_type, AlgorithmDescribeBase)
                else el.default_value
            )
            for el in cls.get_fields()
            if isinstance(el, AlgorithmProperty)
        }


def get_fields_from_algorithm(ald_desc: AlgorithmDescribeBase) -> list[typing.Union[AlgorithmProperty, str]]:
    if ald_desc.__new_style__:
        return base_model_to_algorithm_property(ald_desc.__argument_class__)
    return ald_desc.get_fields()


def is_static(fun):
    if fun is None:
        return False
    args = inspect.getfullargspec(fun).args
    return True if len(args) == 0 else args[0] != "self"


AlgorithmType = typing.TypeVar("AlgorithmType", bound=type[AlgorithmDescribeBase])


class Register(dict, typing.Generic[AlgorithmType]):
    """
    Dict used for register :class:`.AlgorithmDescribeBase` classes.
    All registers from `PartSeg.PartSegCore.register` are this
    :param class_methods: list of method which should be implemented as class method it will be checked during add
    as args or with :meth:`.Register.register`  method
    :param methods: list of method which should be instance method
    """

    def __init__(self, *args: AlgorithmType, class_methods=None, methods=None, suggested_base_class=None, **kwargs):
        """
        :param class_methods: list of method which should be class method
        :param methods: list of method which should be instance method
        :param kwargs: elements passed to OrderedDict constructor (may be initial elements). I suggest to not use this.
        """
        super().__init__(**kwargs)
        self.suggested_base_class = suggested_base_class
        self.class_methods = (
            list(class_methods) if class_methods else getattr(suggested_base_class, "need_class_method", [])
        )
        self.methods = list(methods) if methods else getattr(suggested_base_class, "need_method", [])
        self._old_mapping = {}
        for el in args:
            self.register(el)

    def values(self) -> typing.Iterable[AlgorithmType]:
        # noinspection PyTypeChecker
        return typing.cast("typing.Iterable[AlgorithmType]", super().values())

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and isinstance(other, Register)
            and self.class_methods == other.class_methods
            and self.methods == other.methods
            and self.suggested_base_class == other.suggested_base_class
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, item) -> AlgorithmType:
        # FIXME add better strategy to get proper class when there is conflict of names
        try:
            return typing.cast("AlgorithmType", super().__getitem__(item))
        except KeyError:
            return typing.cast("AlgorithmType", super().__getitem__(self._old_mapping[item]))

    def __contains__(self, item):
        return super().__contains__(item) or item in self._old_mapping

    def register(self, value: AlgorithmType, replace: bool = False, old_names: typing.Optional[list[str]] = None):
        """
        Function for registering :class:`.AlgorithmDescribeBase` based algorithms
        :param value: algorithm to register
        :param replace: replace existing algorithm, be patient with
        :param old_names: list of old names for registered class
        """
        self.check_function(value, "get_name", True)
        try:
            name = value.get_name()
        except (NotImplementedError, AttributeError):
            raise ValueError(f"Class {value} need to implement classmethod 'get_name'") from None
        if name in self and not replace:
            raise ValueError(
                f"Object {self[name]} with this name: '{name}' already exist and register is not in replace mode"
            )
        if not isinstance(name, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(name)}")
        self[name] = value
        if old_names is not None:
            # FIXME add better strategy to get proper class when there is conflict of names
            for old_name in old_names:
                if old_name in self._old_mapping and not replace:
                    raise ValueError(
                        f"Old value mapping for name '{old_name}' already registered."
                        f" Currently pointing to {self._old_mapping[old_name]}"
                    )
                self._old_mapping[old_name] = name
        return value

    @staticmethod
    def check_function(ob, function_name, is_class):
        fun = getattr(ob, function_name, None)
        if not is_class and not inspect.isfunction(fun):
            raise ValueError(f"Class {ob} need to define method {function_name}")
        if is_class and not inspect.ismethod(fun) and not is_static(fun):
            raise ValueError(f"Class {ob} need to define classmethod '{function_name}'")

    def __setitem__(self, key: str, value: AlgorithmType):
        if not issubclass(value, AlgorithmDescribeBase):
            raise ValueError(
                f"Class {value} need to be subclass of {AlgorithmDescribeBase.__module__}.AlgorithmDescribeBase"
            )
        self.check_function(value, "get_name", True)
        self.check_function(value, "get_fields", True)
        try:
            val = value.get_name()
        except (NotImplementedError, AttributeError):
            raise ValueError(f"Class {value} need to implement classmethod 'get_name'") from None
        if not isinstance(val, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(val)}")
        if key != val:
            raise ValueError("Object need to be registered under name returned by get_name function")
        if not value.__new_style__:
            try:
                val = value.get_fields()
                if not isinstance(val, list):
                    raise ValueError(f"Function get_fields of class {value} need return list not {type(val)}")
            except NotImplementedError:
                raise ValueError(f"Method get_fields of class {value} need to be implemented") from None
        for el in self.class_methods:
            self.check_function(value, el, True)
        for el in self.methods:
            self.check_function(value, el, False)

        super().__setitem__(key, value)

    def get_default(self) -> str:
        """
        Calculate default algorithm name for given dict.

        :return: name of algorithm
        """
        try:
            return next(iter(self.keys()))
        except StopIteration:
            raise ValueError("Register does not contain any algorithm.") from None


class AddRegisterMeta(ModelMetaclass):
    def __new__(cls, name, bases, attrs, **kwargs):
        methods = kwargs.pop("methods", [])
        suggested_base_class = kwargs.pop("suggested_base_class", None)
        class_methods = kwargs.pop("class_methods", [])
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)  # pylint: disable=too-many-function-args
        cls2.__register__ = Register(
            class_methods=class_methods, methods=methods, suggested_base_class=suggested_base_class
        )
        return cls2

    def __getitem__(self, item) -> AlgorithmType:
        return self.__register__[item]

    def __contains__(self, item) -> bool:
        return self.__register__.__contains__(item)

    def get(self, item, default=None):
        return self.__register__.get(item, default)


class AlgorithmSelection(BaseModel, metaclass=AddRegisterMeta):  # pylint: disable=E1139
    """
    Base class for algorithm selection.
    For given algorithm there should be Register instance set __register__ class variable.
    """

    name: str
    values: typing.Union[dict[str, typing.Any], PydanticBaseModel] = Field(..., union_mode="left_to_right")
    class_path: str = ""
    if typing.TYPE_CHECKING:
        __register__: Register

    if parse_version(version("pydantic")) < parse_version("2"):

        class Config:
            smart_union = True

    @validator("name")
    def check_name(cls, v):
        if v not in cls.__register__:
            raise ValueError(f"Missed algorithm {v}")
        return v

    @validator("class_path", always=True)
    def update_class_path(cls, v, values):
        if v or "name" not in values:
            return v
        klass = cls.__register__[values["name"]]
        return class_to_str(klass)

    @field_serializer("values")
    def val_serializer(self, value, _info):
        if isinstance(value, PydanticBaseModel):
            return value.dict()
        return value

    @validator("values", pre=True)
    def update_values(cls, v, values):
        # FIXME add better strategy to get proper class when there is conflict of names
        if "name" not in values or not isinstance(v, dict):
            return v
        klass = cls.__register__[values["name"]]
        if not klass.__new_style__ or not klass.__argument_class__.__fields__:
            return v

        dkt_migrated = REGISTER.migrate_data(class_to_str(klass.__argument_class__), {}, v)
        return klass.__argument_class__(**dkt_migrated)

    @classmethod
    def register(
        cls, value: AlgorithmType, replace=False, old_names: typing.Optional[list[str]] = None
    ) -> AlgorithmType:
        """
        Function for registering :class:`.AlgorithmDescribeBase` based algorithms
        :param value: algorithm to register
        :param replace: replace existing algorithm, be patient with
        :param old_names: list of old names for registered class
        """
        return cls.__register__.register(value, replace=replace, old_names=old_names)

    @classmethod
    def get_default(cls):
        name = cls.__register__.get_default()
        return cls(name=name, values=cls[name].get_default_values())

    def algorithm(self):
        return self.__register__[self.name]


class ROIExtractionProfileMeta(ModelMetaclass):
    def __new__(cls, name, bases, attrs, **kwargs):
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)  # pylint: disable=too-many-function-args

        def allow_positional_args(func):
            @wraps(func)
            def _wraps(self, *args, **kwargs):
                if args:
                    warnings.warn(
                        "Positional arguments are deprecated, use keyword arguments instead",
                        FutureWarning,
                        stacklevel=2,
                    )
                    kwargs.update(dict(zip(self.__fields__, args)))
                return func(self, **kwargs)

            return _wraps

        cls2.__init__ = allow_positional_args(cls2.__init__)
        return cls2


class ROIExtractionProfile(BaseModel, metaclass=ROIExtractionProfileMeta):  # pylint: disable=E1139
    """
    :ivar str ~.name: name for segmentation profile
    :ivar str ~.algorithm: Name of algorithm
    :ivar dict ~.values: algorithm parameters
    """

    name: str
    algorithm: str
    values: typing.Any

    @validator("values")
    def validate_values(cls, v, values):  # pylint: disable=no-self-use
        if not isinstance(v, dict):
            return v
        if "algorithm" not in values:
            return v
        from PartSegCore.analysis import AnalysisAlgorithmSelection
        from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection

        name = values["algorithm"]
        is_analysis = name in AnalysisAlgorithmSelection
        is_mask = name in MaskAlgorithmSelection
        if is_analysis == is_mask:
            return v
        algorithm = AnalysisAlgorithmSelection[name] if is_analysis else MaskAlgorithmSelection[name]
        if not algorithm.__new_style__:
            return v
        dkt_migrated = REGISTER.migrate_data(class_to_str(algorithm.__argument_class__), {}, v)
        return algorithm.__argument_class__(**dkt_migrated)

    def pretty_print(self, algorithm_dict):
        if isinstance(algorithm_dict, AlgorithmSelection):
            algorithm_dict = algorithm_dict.__register__
        try:
            algorithm = algorithm_dict[self.algorithm]
        except KeyError:
            return f"{self}\n "
        values = self.values if isinstance(self.values, dict) else self.values.dict()
        if self.name in {"", "Unknown"}:
            return (
                "ROI extraction profile\nAlgorithm: "
                + self.algorithm
                + "\n"
                + self._pretty_print(values, algorithm.get_fields_dict())
            )
        return (
            f"ROI extraction profile name: {self.name}\nAlgorithm: {self.algorithm}\n"
            f"{self._pretty_print(values, algorithm.get_fields_dict())}"
        )

    @classmethod
    def _pretty_print(cls, values: MutableMapping, translate_dict: dict[str, AlgorithmProperty], indent=0):
        if not isinstance(values, MutableMapping):
            return textwrap.indent(str(values), " " * indent)
        res = ""
        for k, v in values.items():
            if k not in translate_dict:
                if isinstance(v, MutableMapping):
                    res += " " * indent + f"{k}: {cls._pretty_print(v, {}, indent + 2)}\n"
                else:
                    res += " " * indent + f"{k}: {v}\n"
                continue
            desc = translate_dict[k]
            res += " " * indent + desc.user_name + ": "
            if issubclass(desc.value_type, Channel) and not isinstance(v, Channel):
                res += str(Channel(v))
            elif issubclass(desc.value_type, AlgorithmDescribeBase):
                if isinstance(v, AlgorithmSelection):
                    name = v.name
                    values_ = v.values
                else:
                    name = v["name"]
                    values_ = v["values"]
                res += desc.possible_values[name].get_name()
                if values_:
                    res += "\n"
                    res += cls._pretty_print(values_, desc.possible_values[name].get_fields_dict(), indent + 2)
            elif isinstance(v, MutableMapping):
                res += cls._pretty_print(v, {}, indent + 2)
            else:
                res += str(v)
            res += "\n"
        return res[:-1]

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.name == other.name
            and self.algorithm == other.algorithm
            and self.values == other.values
        )


def _next_after(type_, value, inf):
    if issubclass(type_, int):
        if inf == math.inf:
            return value + 1
        return value - 1
    if hasattr(math, "nextafter"):
        # TODO fix after drop python 3.8
        return math.nextafter(value, inf)
    return value


def _next_after_with_none(type_, value, inf):
    if value is None:
        return None
    return _next_after(type_, value, inf)


def _calc_value_range(field_info: FieldInfo):
    if field_info.metadata is None:
        return (0, 1000)

    import annotated_types as at

    value_range = (0, 1000)
    for el in field_info.metadata:
        if isinstance(el, at.Ge):
            value_range = el.ge, value_range[1]
        elif isinstance(el, at.Gt):
            value_range = _next_after(field_info.annotation, el.gt, math.inf), value_range[1]
        elif isinstance(el, at.Le):
            value_range = value_range[0], el.le
        elif isinstance(el, at.Lt):
            value_range = value_range[0], _next_after(field_info.annotation, el.lt, -math.inf)
    return value_range


def _field_to_algorithm_property_pydantic_2(name: str, field_info: FieldInfo):
    user_name = field_info.title

    value_range = None
    possible_values = None

    value_type = field_info.annotation
    default_value = field_info.default
    help_text = field_info.description
    if user_name is None:
        user_name = name.replace("_", " ").capitalize()
    if not hasattr(value_type, "__origin__"):
        if issubclass(value_type, (int, float)):
            value_range = _calc_value_range(field_info)

        if issubclass(field_info.annotation, AlgorithmSelection):
            value_type = AlgorithmDescribeBase
            if isinstance(field_info.default, UndefinedType):
                default_value = field_info.default_factory().name
            else:
                default_value = field_info.default.name
            possible_values = field_info.annotation.__register__

    extra = field_info.json_schema_extra or {}

    return AlgorithmProperty(
        name=name,
        user_name=user_name,
        default_value=default_value,
        options_range=value_range,
        value_type=value_type,
        possible_values=possible_values,
        help_text=help_text,
        mgi_options=extra.get("options", {}),
    )


def _field_to_algorithm_property_pydantic_1(name: str, field: "ModelField"):
    user_name = field.field_info.title
    value_range = None
    possible_values = None

    value_type = getattr(field, "annotation", field.type_)
    default_value = field.field_info.default
    help_text = field.field_info.description
    if user_name is None:
        user_name = name.replace("_", " ").capitalize()
    if not hasattr(field.type_, "__origin__"):
        if issubclass(field.type_, (int, float)):
            value_range = (
                field.field_info.ge or _next_after_with_none(field.type_, field.field_info.gt, math.inf) or 0,
                field.field_info.le or _next_after_with_none(field.type_, field.field_info.lt, -math.inf) or 1000,
            )
        if issubclass(field.type_, AlgorithmSelection):
            value_type = AlgorithmDescribeBase
            if isinstance(field.field_info.default, UndefinedType):
                default_value = field.field_info.default_factory().name
            else:
                default_value = field.field_info.default.name
            possible_values = field.type_.__register__

    return AlgorithmProperty(
        name=name,
        user_name=user_name,
        default_value=default_value,
        options_range=value_range,
        value_type=value_type,
        possible_values=possible_values,
        help_text=help_text,
        mgi_options=field.field_info.extra.get("options", {}),
    )


def base_model_to_algorithm_property_pydantic_1(
    obj: type[BaseModel],
) -> list[typing.Union[str, AlgorithmProperty]]:
    """
    Convert pydantic model to list of AlgorithmPropert nad strings.

    :param obj:
    :return:
    """
    res = []
    value: ModelField
    if hasattr(obj, "header") and obj.header():
        res.append(obj.header())
    for name, value in obj.__fields__.items():
        ap = _field_to_algorithm_property(name, value)
        if value.field_info.extra.get("hidden", False):
            continue
        pos = len(res)
        if "position" in value.field_info.extra:
            pos = value.field_info.extra["position"]
        if "prefix" in value.field_info.extra:
            res.insert(pos, value.field_info.extra["prefix"])
            pos += 1

        res.insert(pos, ap)

        if "suffix" in value.field_info.extra:
            res.insert(pos + 1, value.field_info.extra["suffix"])
    return res


def base_model_to_algorithm_property_pydantic_2(
    obj: type[BaseModel],
) -> list[typing.Union[str, AlgorithmProperty]]:
    """
    Convert pydantic model to list of AlgorithmPropert nad strings.

    :param obj:
    :return:
    """
    res = []
    field_info: FieldInfo
    if hasattr(obj, "header") and obj.header():
        res.append(obj.header())
    for name, field_info in obj.__fields__.items():
        ap = _field_to_algorithm_property(name, field_info)
        extra = field_info.json_schema_extra or {}
        if extra.get("hidden", False):
            continue
        pos = len(res)
        if "position" in extra:
            pos = extra["position"]
        if "prefix" in extra:
            res.insert(pos, extra["prefix"])
            pos += 1

        res.insert(pos, ap)

        if "suffix" in extra:
            res.insert(pos + 1, extra["suffix"])
    return res


if version("pydantic") < "2":
    _field_to_algorithm_property = _field_to_algorithm_property_pydantic_1
    base_model_to_algorithm_property = base_model_to_algorithm_property_pydantic_1
else:
    _field_to_algorithm_property = _field_to_algorithm_property_pydantic_2
    base_model_to_algorithm_property = base_model_to_algorithm_property_pydantic_2
